from .context import get_current_config, connections, log, run_task, metric
import time
import datetime
import gevent
import argparse
import random
import shlex
import traceback
from collections import defaultdict
from bson import ObjectId
from redis.lock import LuaLock
from .processes import Process, ProcessPool
from .utils import MovingETA, normalize_command
from .queue import Queue


class Agent(Process):
    """ MRQ Agent manages its local worker pool and takes turns in orchestrating the others in its group. """

    def __init__(self, worker_group=None):
        self.greenlets = {}
        self.id = ObjectId()
        self.worker_group = worker_group or get_current_config()["worker_group"]
        self.pool = ProcessPool(extra_env={
            "MRQ_AGENT_ID": str(self.id),
            "MRQ_WORKER_GROUP": self.worker_group
        })
        self.config = get_current_config()
        self.status = "started"
        metric("agent", data={"worker_group": self.worker_group, "agent_id": self.id})
        self.dateorchestrated = None

        # global redis key used to ensure only one agent orchestrator runs at a time
        self.redis_queuestats_lock_key = "%s:queuestatslock" % (self.config["redis_prefix"])

        # global HSET redis key used to store queue stats
        self.redis_queuestats_key= "%s:queuestats" % (self.config["redis_prefix"])

    def work(self):

        self.install_signal_handlers()
        self.datestarted = datetime.datetime.utcnow()

        self.pool.start()
        self.manage()

        self.greenlets["orchestrate"] = gevent.spawn(self.greenlet_orchestrate)
        self.greenlets["orchestrate"].start()

        self.greenlets["manage"] = gevent.spawn(self.greenlet_manage)
        self.greenlets["manage"].start()

        # Disabled for now
        # self.greenlets["queuestats"] = gevent.spawn(self.greenlet_queuestats)
        # self.greenlets["queuestats"].start()

        try:
            self.pool.wait()
        finally:
            self.shutdown_now()
            self.status = "stop"
            self.manage()

    def shutdown_now(self):
        self.pool.terminate()

        for g in self.greenlets.values():
            g.kill()

    def shutdown_graceful(self):
        self.pool.stop(timeout=None)

    def greenlet_manage(self):
        """ This greenlet always runs in background to update current status
            in MongoDB every N seconds.
         """

        while True:
            try:
                self.manage()
            except Exception as e:  # pylint: disable=broad-except
                log.error("When reporting: %s" % e)
            finally:
                time.sleep(self.config["report_interval"])

    def manage(self):

        report = self.get_agent_report()

        try:
            db = connections.mongodb_jobs.mrq_agents.find_and_modify({
                "_id": ObjectId(self.id)
            }, {"$set": report}, upsert=True)
            if not db:
                return
        except Exception as e:  # pylint: disable=broad-except
            log.debug("Agent report failed: %s" % e)
            return

        # If the desired_workers was changed by an orchestrator, apply the changes locally
        if self.status != "stop" and sorted(db.get("desired_workers", [])) != sorted(self.pool.desired_commands):

            group = self.fetch_worker_group_definition()
            process_termination_timeout = float(group.get("process_termination_timeout") or 60)
            self.pool.set_commands(db.get("desired_workers", []), timeout=process_termination_timeout)

    def get_agent_report(self):
        report = {
            "current_workers": [p["command"] for p in self.pool.processes],
            "total_cpu": get_current_config()["total_cpu"],
            "total_memory": get_current_config()["total_memory"],
            "worker_group": self.worker_group,
            "status": self.status,
            "dateorchestrated": self.dateorchestrated,
            "datestarted": self.datestarted,
            "datereported": datetime.datetime.utcnow(),
            "dateexpires": datetime.datetime.utcnow() + datetime.timedelta(seconds=(self.config["report_interval"] * 3) + 5)
        }
        metric("agent", data={"worker_group": self.worker_group, "agent_id": self.id, "worker_count": len(self.pool.processes)})
        return report

    def greenlet_orchestrate(self):
        while True:
            try:
                self.orchestrate()
            except Exception as e:
                log.error("Orchestration error! %s" % e)
                traceback.print_exc()

            time.sleep(self.config["orchestrate_interval"])

    def orchestrate(self):
        run_task("mrq.basetasks.orchestrator.Orchestrate", {})

    def greenlet_queuestats(self):

        interval = min(self.config["orchestrate_interval"], 1 * 60)
        lock_timeout = 5 * 60 + (interval * 2)

        while True:
            lock = LuaLock(connections.redis, self.redis_queuestats_lock_key,
                           timeout=lock_timeout, thread_local=False, blocking=False)
            with lock:
                lock_expires = time.time() + lock_timeout
                self.queue_etas = defaultdict(lambda: MovingETA(5))

                while True:
                    self.queuestats()

                    # Because queue stats can be expensive, we try to keep the lock on the same agent
                    lock_extend = (time.time() + lock_timeout) - lock_expires
                    lock_expires += lock_extend
                    lock.extend(lock_extend)

                    time.sleep(interval)

            time.sleep(interval)

    def queuestats(self):
        """ Compute ETAs for every known queue & subqueue """

        start_time = time.time()
        log.debug("Starting queue stats...")

        # Fetch all known queues
        queues = [Queue(q) for q in Queue.all_known()]

        new_queues = {queue.id for queue in queues}
        old_queues = set(self.queue_etas.keys())

        for deleted_queue in old_queues.difference(new_queues):
            self.queue_etas.pop(deleted_queue)

        t = time.time()
        stats = {}

        for queue in queues:
            cnt = queue.count_jobs_to_dequeue()
            eta = self.queue_etas[queue.id].next(cnt, t=t)

            # Number of jobs to dequeue, ETA, Time of stats
            stats[queue.id] = "%d %s %d" % (cnt, eta if eta is not None else "N", int(t))

        with connections.redis.pipeline(transaction=True) as pipe:
            if random.randint(0, 100) == 0 or len(stats) == 0:
                pipe.delete(self.redis_queuestats_key)
            if len(stats) > 0:
                pipe.hmset(self.redis_queuestats_key, stats)
            pipe.execute()

        log.debug("... done queue stats in %0.4fs" % (time.time() - start_time))

    def fetch_worker_group_definition(self):
        definition = connections.mongodb_jobs.mrq_workergroups.find_one({"_id": self.worker_group})

        # Prepend all commands by their worker profile.
        commands = []
        for command in definition.get("commands", []):
            simplified_command, worker_count = normalize_command(command, self.worker_group)
            commands.extend([simplified_command] * worker_count)

        definition["commands"] = commands
        return definition



#!/usr/bin/env python
from __future__ import print_function

import os
import sys
is_pypy = '__pypy__' in sys.builtin_module_names

# Needed to make getaddrinfo() work in pymongo on Mac OS X
# Docs mention it's a better choice for Linux as well.
# This must be done asap in the worker

if "GEVENT_RESOLVER" not in os.environ and not is_pypy:
    os.environ["GEVENT_RESOLVER"] = "ares"

# We must still monkey-patch the methods for job sub-pools.
from gevent import monkey
monkey.patch_all()

import argparse
import ujson as json
import json as json_stdlib
import datetime

sys.path.insert(0, os.getcwd())

from mrq import config, utils
from mrq.context import set_current_config, set_logger_config, set_current_job, connections
from mrq.job import queue_job
from mrq.utils import load_class_by_path, MongoJSONEncoder

def main():

    parser = argparse.ArgumentParser(description='Runs a task')

    cfg = config.get_config(parser=parser, config_type="run", sources=("file", "env", "args"))
    cfg["is_cli"] = True
    set_current_config(cfg)
    set_logger_config()

    if len(cfg["taskargs"]) == 1:
        params = json.loads(cfg["taskargs"][0])  # pylint: disable=no-member
    else:
        params = {}

        # mrq-run taskpath a 1 b 2 => {"a": "1", "b": "2"}
        for group in utils.group_iter(cfg["taskargs"], n=2):
            if len(group) != 2:
                print("Number of arguments wasn't even")
                sys.exit(1)
            params[group[0]] = group[1]

    if cfg["queue"]:
        ret = queue_job(cfg["taskpath"], params, queue=cfg["queue"])
        print(ret)
    else:
        worker_class = load_class_by_path(cfg["worker_class"])
        job = worker_class.job_class(None)
        job.set_data({
            "path": cfg["taskpath"],
            "params": params,
            "queue": cfg["queue"]
        })
        job.datestarted = datetime.datetime.utcnow()
        set_current_job(job)
        ret = job.perform()
        print(json_stdlib.dumps(ret, cls=MongoJSONEncoder))  # pylint: disable=no-member

    # This shouldn't be needed as the process will exit and close any remaining sockets
    # connections.redis.connection_pool.disconnect()

if __name__ == "__main__":
    main()



from __future__ import division

from future.builtins import bytes, str, object

import time
from bson import ObjectId
from . import context
from . import job as jobmodule
import binascii
from .redishelpers import redis_key

import sys
from future import standard_library
from itertools import chain

PY3 = sys.version_info > (3,)
standard_library.install_aliases()


class Queue(object):
    """ A Queue for Jobs. """

    is_raw = False
    is_timed = False
    is_sorted = False
    is_set = False

    is_reverse = False

    # root_id will contain the root queue id without any trailing subqueue delimiter
    # e.g. if self.id is "some_queue/" then self.root_id will contain "some_queue"
    # and if self.id is "some_queue/some_subqueue" then self.root_id will contain "some_queue"
    root_id = None
    is_subqueue = False

    use_large_ids = False

    # This is a mutable type so it is shared by all instances
    # of Queue in the current process
    paused_queues = set()

    def __new__(cls, queue_id, **kwargs):
        """ Creates a new instance of the right queue type """

        if cls is not Queue:
            return object.__new__(cls)

        if isinstance(queue_id, Queue):
            queue_id = queue_id.id

        queue_type = Queue.get_queue_type(queue_id)

        if queue_type == "regular":
            from .queue_regular import QueueRegular
            return QueueRegular(queue_id, **kwargs)
        else:
            from .queue_raw import QueueRaw
            return QueueRaw(queue_id, **kwargs)

    def __init__(self, queue_id):

        if queue_id[-8:] == "_reverse":
            self.is_reverse = True
            queue_id = queue_id[:-8]

        self.id = queue_id

        self.root_id = self.id

        if "/" in self.id:
            # Get the root queue id with no trailing delimiter
            self.root_id = self.id.split("/")[0]

            if self.id.split("/")[1]:
                self.is_subqueue = True

        self.use_large_ids = context.get_current_config()["use_large_job_ids"]

    @classmethod
    def get_queue_type(cls, queue_id):
        """ Return the queue type, currently determined only by its suffix. """

        for queue_type in ("timed_set", "sorted_set", "set", "raw"):
            if queue_id.split("/")[0].endswith("_%s" % queue_type):
                return queue_type

        return "regular"

    @classmethod
    def get_queues_config(cls):
        """ Returns the queues configuration dict """
        _config = context.get_current_config()
        return _config.get("raw_queues") or _config.get("queues_config") or {}

    def get_config(self):
        """ Returns the specific configuration for this queue """
        return Queue.get_queues_config().get(self.root_id) or {}

    def serialize_job_ids(self, job_ids):
        """ Returns job_ids serialized for storage in Redis """
        if len(job_ids) == 0 or self.use_large_ids:
            return job_ids
        elif isinstance(job_ids[0], ObjectId):
            return [x.binary for x in job_ids]
        else:
            return [bytes.fromhex(str(x)) for x in job_ids]

    def unserialize_job_ids(self, job_ids):
        """ Unserialize job_ids stored in Redis """
        if len(job_ids) == 0 or self.use_large_ids:
            return job_ids
        else:
            return [binascii.hexlify(x.encode('utf-8') if (PY3 and isinstance(x, str)) else x).decode('ascii')
                    for x in job_ids]

    def _get_pausable_id(self):
      """
          Get the queue id (either id or root_id) that should be used to pause/unpause the current queue
          TODO: handle subqueues with more than one level, e.g. "queue/subqueue/"
      """
      queue = self.id
      if self.id.endswith("/"):
          queue = self.root_id
      return queue

    def pause(self):
        """ Adds this queue to the set of paused queues """
        context.connections.redis.sadd(redis_key("paused_queues"), self._get_pausable_id())

    def is_paused(self):
        """
            Returns wether the queue is paused or not.
            Warning: this does NOT ensure that the queue was effectively added to
            the set of paused queues. See the 'paused_queues_refresh_interval' option.
        """
        root_is_paused = False
        if self.root_id != self.id:
            root_is_paused = context.connections.redis.sismember(redis_key("paused_queues"), self.root_id)

        return root_is_paused or context.connections.redis.sismember(redis_key("paused_queues"), self.id)

    def resume(self):
        """ Resumes a paused queue """
        context.connections.redis.srem(redis_key("paused_queues"), self._get_pausable_id())

    def count_jobs_to_dequeue(self):
        """ Returns the number of jobs that can be dequeued right now from the queue. """

        return self.size()

    @classmethod
    def all_active(cls):
        """ List active queues, based on their lengths in Redis. Warning, uses the unscalable KEYS redis command """

        prefix = context.get_current_config()["redis_prefix"]
        queues = []
        for key in context.connections.redis.keys():
            if key.startswith(prefix):
                queues.append(Queue(key[len(prefix) + 3:]))

        return queues

    @classmethod
    def all_known(cls, sources=None, prefixes=None):
        """ List all currently known queues """

        sources = sources or ("config", "jobs", "raw_subqueues")

        queues = set()

        if "config" in sources and not prefixes:
            # Some queues are explicitly declared in the config (including all root raw queues)
            cfg = context.get_current_config()

            queues_from_config = [
                t.get("queue")
                for t in (cfg.get("tasks") or {}).values()
                if t.get("queue")
            ]

            queues_from_config += Queue.get_queues_config().keys()

            queues_from_config += [
                t.get("retry_queue")
                for t in Queue.get_queues_config().values()
                if t.get("retry_queue")
            ]

            queues |= set(queues_from_config)

        if "jobs" in sources:

            # This will get all queues from mongodb, including those where we have only non-queued jobs
            for q in context.connections.mongodb_jobs.mrq_jobs.distinct("queue"):
                if prefixes and not any(q.startswith(p) for p in prefixes):
                    continue
                queues.add(q)

        if "raw_subqueues" in sources:
            for q in Queue.get_queues_config():
                if prefixes and not any(q + "/" == p for p in prefixes):
                    continue
                queue_obj = Queue(q)
                if queue_obj.is_raw and queue_obj.has_subqueues:
                    # TODO: optimize this with a single SUNION on all keys
                    queues |= queue_obj.get_known_subqueues()

        return queues

    @classmethod
    def all(cls):
        """ List all queues in MongoDB via aggregation, with their queued jobs counts. Might be slow. """

        # Start with raw queues we know exist from the config
        queues = {x: 0 for x in Queue.get_queues_config()}

        stats = list(context.connections.mongodb_jobs.mrq_jobs.aggregate([
            {"$match": {"status": "queued"}},
            {"$group": {"_id": "$queue", "jobs": {"$sum": 1}}}
        ]))

        queues.update({x["_id"]: x["jobs"] for x in stats})

        return queues

    def use_notify(self):
        """ Does this queue use notifications? """
        return bool(self.get_config().get("notify"))

    def notify(self, new_jobs_count):
        """ We just queued new_jobs_count jobs on this queue, wake up the workers if needed """

        if not self.use_notify():
            return

        # Not really useful to send more than 100 notifs (to be configured)
        count = min(new_jobs_count, 100)

        notify_key = redis_key("notify", self)

        context.connections.redis.lpush(notify_key, *([1] * count))
        context.connections.redis.expire(notify_key, max(1, int(context.get_current_config()["max_latency"] * 2)))


#
# Deprecated methods. Tagged for removal in 1.0.0
#

def send_raw_tasks(*args, **kwargs):
    return jobmodule.queue_raw_jobs(*args, **kwargs)


def send_task(path, params, **kwargs):
    return send_tasks(path, [params], **kwargs)[0]


def send_tasks(path, params_list, queue=None, sync=False, batch_size=1000):
    if sync:
        return [context.run_task(path, params) for params in params_list]

    return jobmodule.queue_jobs(path, params_list, queue=queue, batch_size=batch_size)



from future.builtins import str
from mrq.queue import Queue
from mrq.task import Task
from mrq.job import Job
from mrq.context import log, connections, run_task, get_current_config, subpool_map
from collections import defaultdict
import math
import shlex
import argparse
from ..config import add_parser_args
from ..utils import normalize_command
import traceback
import datetime
import re


class Orchestrate(Task):

    max_concurrency = 1

    def run(self, params):

        self.config = get_current_config()

        concurrency = int(params.get("concurrency", 5))
        groups = self.fetch_worker_group_definitions()
        if len(groups) == 0:
            log.error("No worker group definition yet. Can't orchestrate!")
            return

        subpool_map(concurrency, self.orchestrate, groups)

    def redis_orchestrator_lock_key(self, worker_group):
        """ Returns the global redis key used to ensure only one agent orchestrator runs at a time """
        return "%s:orchestratorlock:%s" % (get_current_config()["redis_prefix"], worker_group)

    def orchestrate(self, worker_group):
        try:
            self.do_orchestrate(worker_group)
        except Exception as e:
            log.error("Orchestration error! %s" % e)
            traceback.print_exc()

    def do_orchestrate(self, group):
        """ Manage the desired workers of *all* the agents in the given group """

        log.debug("Starting orchestration run for worker group %s" % group["_id"])

        agents = self.fetch_worker_group_agents(group)

        # Evaluate what workers are currently, rightfully there. They won't be touched.
        for agent in agents:
            desired_workers = self.get_desired_workers_for_agent(group, agent)
            agent["new_desired_workers"] = []
            agent["new_desired_workers"] = desired_workers

        for agent in agents:
            if sorted(agent["new_desired_workers"]) != sorted(agent.get("desired_workers", [])):
                connections.mongodb_jobs.mrq_agents.update_one({"_id": agent["_id"]}, {"$set": {
                    "desired_workers": agent["new_desired_workers"]
                }})

        # Remember the date of the last successful orchestration (will be reported)
        self.dateorchestrated = datetime.datetime.utcnow()

        log.debug("Orchestration finished.")

    def redis_queuestats_key(self):
        """ Returns the global HSET redis key used to store queue stats """
        return "%s:queuestats" % (get_current_config()["redis_prefix"])

    def get_desired_workers_for_agent(self, group, agent):
        return group.get("commands", [])

    def fetch_worker_group_reports(self, worker_group, projection=None):
        return list(connections.mongodb_jobs.mrq_workers.find({
            "config.worker_group": worker_group["_id"]
        }, projection=projection))

    def fetch_worker_group_definitions(self):

        definitions = list(connections.mongodb_jobs.mrq_workergroups.find())

        for definition in definitions:
            commands = []
            # Prepend all commands by their worker group.
            for command in definition.get("commands", []):
                simplified_command, worker_count = normalize_command(command, definition["_id"])
                commands.extend([simplified_command] * worker_count)
            definition["commands"] = commands

        return definitions

    def fetch_worker_group_agents(self, worker_group):
        return list(connections.mongodb_jobs.mrq_agents.find({"worker_group": worker_group["_id"], "status": "started"}))

    def get_config_for_profile(self, profile):
        parser = argparse.ArgumentParser()
        add_parser_args(parser, "worker")
        parts = shlex.split(profile["command"])
        if "mrq-worker" in parts:
            parts = parts[parts.index("mrq-worker") + 1:]
        return parser.parse_args(parts)



from __future__ import print_function
from future.builtins import str
import argparse
import os
import sys
import re
import psutil
from .version import VERSION
from .utils import get_local_ip
import atexit
import logging


def add_parser_args(parser, config_type):

    # General arguments

    parser.add_argument(
        '--trace_greenlets',
        action='store_true',
        default=False,
        help='Collect stats about each greenlet execution time and switches.')

    parser.add_argument(
        '--trace_memory',
        action='store_true',
        default=False,
        help='Collect stats about memory for each task. Incompatible with --greenlets > 1')

    parser.add_argument(
        '--trace_io',
        action='store_true',
        default=False,
        help='Collect stats about all I/O operations')

    parser.add_argument(
        '--print_mongodb',
        action='store_true',
        default=False,
        help='Print all MongoDB requests')

    parser.add_argument(
        '--trace_memory_type',
        action='store',
        default="",
        help='Create a .png object graph in trace_memory_output_dir ' +
             'with a random object of this type.')

    parser.add_argument(
        '--trace_memory_output_dir',
        action='store',
        default="memory_traces",
        help='Directory where to output .pngs with object graphs')

    parser.add_argument(
        '--profile',
        action='store_true',
        default=False,
        help='Run profiling on the whole worker')

    parser.add_argument(
        '--mongodb_jobs', '--mongodb',
        action='store',
        default="mongodb://127.0.0.1:27017/mrq",
        help='MongoDB URI for the jobs, scheduled_jobs & workers database')

    parser.add_argument(
        '--mongodb_logs',
        action='store',
        default="1",
        help='MongoDB URI for the logs database. ' +
             ' "0" will disable remote logs, "1" will use main MongoDB.')

    parser.add_argument(
        '--redis',
        action='store',
        default="redis://127.0.0.1:6379",
        help='Redis URI')

    parser.add_argument(
        '--redis_prefix',
        action='store',
        default="mrq",
        help='Redis key prefix')

    parser.add_argument(
        '--redis_max_connections',
        action='store',
        type=int,
        default=1000,
        help='Redis max connection pool size')

    parser.add_argument(
        '--redis_timeout',
        action='store',
        type=float,
        default=30,
        help='Redis connection pool timeout to wait for an available connection')

    parser.add_argument(
        '--name',
        default=None,
        action='store',
        help='Specify a different name')

    parser.add_argument(
        '--config',
        '-c',
        default=None,
        action="store",
        help='Path of a config file')

    parser.add_argument(
        '--worker_class',
        default="mrq.worker.Worker",
        action="store",
        help='Path to a custom worker class')

    parser.add_argument(
        '--quiet',
        default=False,
        action='store_true',
        help='Don\'t output task logs')

    parser.add_argument(
        '--queue_ttl',
        default=24 * 60 * 60,
        action='store',
        help='Seconds the Queue size is stored in redis'
    )

    parser.add_argument(
        '--log_handler',
        default="mrq.logger.MongoHandler",
        action="store",
        help='Path to a log handler class')

    parser.add_argument(
        '--log_format',
        default="%(asctime)s [%(levelname)s] %(message)s",
        action="store",
        help='log format')

    parser.add_argument(
        '--log_level',
        default="DEBUG",
        action="store",
        help='set the logging level')

    parser.add_argument(
        '--version',
        '-v',
        default=False,
        action="store_true",
        help='Prints current MRQ version')

    parser.add_argument(
        '--no_import_patch',
        default=False,
        action='store_true',
        help='(DEPRECATED) Skips patching __import__ to fix gevent bug #108')

    parser.add_argument(
        '--add_network_latency',
        default="0",
        action='store',
        type=str,
        help='Adds random latency to the network calls, zero to N seconds. Can be a range (1-2)')

    parser.add_argument(
        '--default_job_ttl',
        default=180 * 24 * 3600,
        action='store',
        type=float,
        help='Seconds the tasks are kept in MongoDB when statuses are not success, abort, cancel and started')

    parser.add_argument(
        '--default_job_result_ttl',
        default=7 * 24 * 3600,
        action='store',
        type=float,
        help='Seconds the results are kept in MongoDB when status is success')

    parser.add_argument(
        '--default_job_abort_ttl',
        default=24 * 3600,
        action='store',
        type=float,
        help='Seconds the tasks are kept in MongoDB when status is abort')

    parser.add_argument(
        '--default_job_cancel_ttl',
        default=24 * 3600,
        action='store',
        type=float,
        help='Seconds the tasks are kept in MongoDB when status is cancel')

    parser.add_argument(
        '--default_job_timeout',
        default=3600,
        action='store',
        type=float,
        help='In seconds, delay before interrupting the job')

    parser.add_argument(
        '--default_job_max_retries',
        default=3,
        action='store',
        type=int,
        help='Set the status to "maxretries" after retrying that many times')

    parser.add_argument(
        '--default_job_retry_delay',
        default=3,
        action='store',
        type=int,
        help='Seconds before a job in retry status is requeued again')

    parser.add_argument(
        '--use_large_job_ids',
        action='store_true',
        default=False,
        help='Do not use compacted job IDs in Redis. For compatibility with 0.1.x only')

    # mrq-run-specific arguments

    if config_type == "run":

        parser.add_argument(
            '--queue',
            action='store',
            default="",
            help='Queue the task on this queue instead of running it right away')

        parser.add_argument(
            'taskpath',
            action='store',
            help='Task to run')

        parser.add_argument(
            'taskargs',
            action='store',
            default='{}',
            nargs='*',
            help='JSON-encoded arguments, or "key value" pairs')

    # Dashboard-specific arguments

    elif config_type == "dashboard":

        parser.add_argument(
            '--dashboard_httpauth',
            default="",
            action="store",
            help='HTTP Auth for the Dashboard. Format is user:pass')

        parser.add_argument(
            '--dashboard_queue',
            default=None,
            action="store",
            help='Default queue for dashboard actions.')

        parser.add_argument(
            '--dashboard_port',
            default=5555,
            action="store",
            type=int,
            help='Use this port for mrq-dashboard. 5555 by default.')

        parser.add_argument(
            '--dashboard_ip',
            default="0.0.0.0",
            action="store",
            type=str,
            help='Bind the dashboard to this IP. Default is "0.0.0.0", use "127.0.0.1" to restrict access.')

    # Agent-specific args
    elif config_type == "agent":

        parser.add_argument(
            '--worker_group',
            default="default",
            action="store",
            type=str,
            help='The name of the worker group to manage')

        parser.add_argument(
            '--total_memory',
            default=int(psutil.virtual_memory().total * 0.8 / (1024 * 1024)),
            action="store",
            type=int,
            help="How much memory MB this agent's workers can use. Used for scheduling, not a hard limit.")

        parser.add_argument(
            '--total_cpu',
            default=psutil.cpu_count(logical=True) * 1024,
            action="store",
            type=int,
            help="How much CPU units this agent's workers can use. We recommend using 1024 per CPU.")

        parser.add_argument(
            '--orchestrate_interval',
            default=30,
            action="store",
            type=float,
            help="How much seconds to wait between orchestration runs.")

        parser.add_argument(
            '--report_interval',
            default=10,
            action='store',
            type=float,
            help='Seconds between agent reports to MongoDB')

        parser.add_argument(
            '--autoscaling_taskpath',
            default=None,
            action='store',
            type=str,
            help='Path to a task that can perform autoscaling actions during orchestration')

    # Worker-specific args
    elif config_type == "worker":

        parser.add_argument(
            '--max_jobs',
            default=0,
            type=int,
            action='store',
            help='Gevent: max number of jobs to do before quitting.' +
                 ' Temp workaround for memory leaks')

        parser.add_argument(
            '--max_time',
            default=0.0,
            type=float,
            action='store',
            help='Max time a worker should run before quitting.')

        parser.add_argument(
            '--max_memory',
            default=0,
            type=int,
            action='store',
            help='Max memory (in Mb) after which the process will be shut down. Use with mrq-agent' +
                 'to automatically respawn the worker when this happens')

        parser.add_argument(
            '--greenlets',
            '--gevent',  # deprecated
            '-g',
            default=1,
            type=int,
            action='store',
            help='Max number of greenlets to use')

        parser.add_argument(
            '--processes',
            '-p',
            default=0,
            type=int,
            action='store',
            help='Number of processes to launch')

        parser.add_argument(
            '--scheduler',
            default=False,
            action='store_true',
            help='Run the scheduler')

        parser.add_argument(
            '--ensure_indexes',
            default=False,
            action='store_true',
            help='Ensures the internal MongoDB indexes of MRQ are built, or does so in the background')

        parser.add_argument(
            '--scheduler_interval',
            default=60,
            action='store',
            type=float,
            help='Seconds between scheduler checks')

        parser.add_argument(
            '--report_interval',
            default=10,
            action='store',
            type=float,
            help='Seconds between worker reports to MongoDB')

        parser.add_argument(
            '--report_file',
            default="",
            action='store',
            type=str,
            help='Filepath of a json dump of the worker status. Disabled if none')

        parser.add_argument(
            '--agent_id',
            default="",
            action='store',
            type=str,
            help='ID of the Agent this worker process is linked to')

        parser.add_argument(
            '--worker_id',
            default="",
            action='store',
            type=str,
            help='ID of this worker process. Should be left empty to be autogenerated in most cases.')

        parser.add_argument(
            '--worker_group',
            default="",
            action='store',
            type=str,
            help='Worker group of the agent this worker was launched from')

        parser.add_argument(
            '--task_whitelist',
            default="",
            action='store',
            type=str,
            help='Comma-separated list of task paths to do exclusively on this worker. Only for regular queues.')

        parser.add_argument(
            '--task_blacklist',
            default="",
            action='store',
            type=str,
            help='Comma-separated list of task paths to do exclude from this worker. Only for regular queues.')

        parser.add_argument(
            'queues',
            nargs='*',
            default=["default"],
            help='The queues to listen on (default: \'default\')')

        parser.add_argument(
            '--subqueues_refresh_interval',
            default=60,
            action='store',
            type=float,
            help="Seconds between worker refreshes of the known subqueues")

        parser.add_argument(
            '--paused_queues_refresh_interval',
            default=10,
            action='store',
            type=float,
            help="Seconds between worker refreshes of the paused queues list")

        parser.add_argument(
            '--admin_port',
            default=0,
            action="store",
            type=int,
            help='Start an admin server on this port, if provided. Incompatible with --processes')

        parser.add_argument(
            '--admin_ip',
            default="127.0.0.1",
            action="store",
            type=str,
            help='IP for the admin server to listen on. Use "0.0.0.0" to allow access from outside')

        parser.add_argument(
            '--local_ip',
            default=get_local_ip(),
            action="store",
            type=str,
            help='Overwrite the local IP, to be displayed in the dashboard.')

        parser.add_argument(
            '--external_ip',
            default=None,
            action="store",
            type=str,
            help='Overwrite the external IP, to be displayed in the dashboard.')

        parser.add_argument(
            '--max_latency',
            default=1.,
            type=float,
            action='store',
            help='Max seconds while worker may sleep waiting for a new job. ' +
                 'Can be < 1.')

        parser.add_argument(
            '--dequeue_strategy',
            default="sequential",
            type=str,
            action='store',
            help='Strategy for dequeuing multiple queues. Default is \'sequential\',' +
                 'to dequeue them in command-line order.')


class ArgumentParserIgnoringDefaults(argparse.ArgumentParser):
    def add_argument(self, *args, **kwargs):
        kwargs.pop("default", None)
        return argparse.ArgumentParser.add_argument(self, *args, **kwargs)


def get_config(
        sources=(
            "file",
            "env"),
        env_prefix="MRQ_",
        file_path=None,
        parser=None,
        extra=None,
        config_type=None):
    """ Returns a config dict merged from several possible sources """

    if not parser:
        parser = argparse.ArgumentParser()

    add_parser_args(parser, config_type)
    parser_types = {action.dest: action.type for action in parser._actions if action.dest}

    if config_type in ["run"]:
        default_config = parser.parse_args(["notask"]).__dict__
    else:
        default_config = parser.parse_args([]).__dict__

    # Keys that can't be passed from the command line
    default_config["tasks"] = {}
    default_config["log_handlers"] = {}
    default_config["scheduled_tasks"] = {}

    # Only keep values actually passed on the command line
    from_args = {}
    if "args" in sources:
        cmdline_parser = ArgumentParserIgnoringDefaults(argument_default=argparse.SUPPRESS)
        add_parser_args(cmdline_parser, config_type)
        from_args = cmdline_parser.parse_args().__dict__

    # If we were given another config file, use it

    if file_path is not None:
        config_file = file_path
    elif from_args.get("config"):
        config_file = from_args.get("config")
    # If a mrq-config.py file is in the current directory, use it!
    elif os.path.isfile(os.path.join(os.getcwd(), "mrq-config.py")):
        config_file = os.path.join(os.getcwd(), "mrq-config.py")
    else:
        config_file = None

    from_file = {}
    if config_file and "file" in sources:
        sys.path.insert(0, os.path.dirname(config_file))
        config_module = __import__(os.path.basename(config_file.replace(".py", "")))
        sys.path.pop(0)
        for k, v in config_module.__dict__.items():

            # We only keep variables starting with an uppercase character.
            if k[0].isupper():
                from_file[k.lower()] = v

    # Merge the config in the order given by the user
    merged_config = default_config

    config_keys = set(list(default_config.keys()) + list(from_file.keys()))

    for part in sources:
        for name in config_keys:

            if part == "env":
                value = os.environ.get(env_prefix + name.upper())
                if value:
                    if name == "queues":
                        value = re.split("\s+", value)
                    if parser_types.get(name):
                        value = parser_types[name](value)
                    merged_config[name] = value
            elif part == "args" and name in from_args:
                merged_config[name] = from_args[name]
            elif part == "file" and name in from_file:
                merged_config[name] = from_file[name]

    if extra:
        merged_config.update(extra)

    if merged_config["profile"]:
        import cProfile, pstats
        profiler = cProfile.Profile()
        profiler.enable()

        def print_profiling():
            pstats.Stats(profiler).sort_stats("cumulative").print_stats()

        atexit.register(print_profiling)

    if merged_config["version"]:
        print("MRQ version: %s" % VERSION)
        print("Python version: %s" % sys.version)
        sys.exit(1)

    if "no_import_patch" in from_args:
        print("WARNING: --no_import_patch will be deprecated in MRQ 1.0!")

    return merged_config



from itertools import count as itertools_count
import traceback
import time
import gevent


def subpool_map(pool_size, func, iterable):
    """ Starts a Gevent pool and run a map. Takes care of setting current_job and cleaning up. """

    from .context import get_current_job, set_current_job, log

    if not pool_size:
        return [func(*args) for args in iterable]

    counter = itertools_count()

    current_job = get_current_job()

    def inner_func(*args):
        """ As each call to 'func' will be done in a random greenlet of the subpool, we need to
            register their IDs with set_current_job() to make get_current_job() calls work properly
            inside 'func'.
        """
        next(counter)
        if current_job:
            set_current_job(current_job)

        try:
          ret = func(*args)
        except Exception as exc:
          trace = traceback.format_exc()
          exc.subpool_traceback = trace
          raise

        if current_job:
            set_current_job(None)
        return ret

    def inner_iterable():
        """ This will be called inside the pool's main greenlet, which ID also needs to be registered """
        if current_job:
            set_current_job(current_job)

        for x in iterable:
            yield x

        if current_job:
            set_current_job(None)

    start_time = time.time()
    pool = gevent.pool.Pool(size=pool_size)
    ret = pool.map(inner_func, inner_iterable())
    pool.join(raise_error=True)
    total_time = time.time() - start_time

    log.debug("SubPool ran %s greenlets in %0.6fs" % (counter, total_time))

    return ret


def subpool_imap(pool_size, func, iterable, flatten=False, unordered=False, buffer_size=None):
  """ Generator version of subpool_map. Should be used with unordered=True for optimal performance """

  from .context import get_current_job, set_current_job, log

  if not pool_size:
    for args in iterable:
      yield func(*args)

  counter = itertools_count()

  current_job = get_current_job()

  def inner_func(*args):
    """ As each call to 'func' will be done in a random greenlet of the subpool, we need to
        register their IDs with set_current_job() to make get_current_job() calls work properly
        inside 'func'.
    """
    next(counter)
    if current_job:
      set_current_job(current_job)

    try:
      ret = func(*args)
    except Exception as exc:
      trace = traceback.format_exc()
      exc.subpool_traceback = trace
      raise

    if current_job:
      set_current_job(None)
    return ret

  def inner_iterable():
    """ This will be called inside the pool's main greenlet, which ID also needs to be registered """
    if current_job:
      set_current_job(current_job)

    for x in iterable:
      yield x

    if current_job:
      set_current_job(None)

  start_time = time.time()
  pool = gevent.pool.Pool(size=pool_size)

  if unordered:
    iterator = pool.imap_unordered(inner_func, inner_iterable(), maxsize=buffer_size or pool_size)
  else:
    iterator = pool.imap(inner_func, inner_iterable())

  for x in iterator:
    if flatten:
      for y in x:
        yield y
    else:
      yield x

  pool.join(raise_error=True)
  total_time = time.time() - start_time

  log.debug("SubPool ran %s greenlets in %0.6fs" % (counter, total_time))



#!/usr/bin/env python
import os
import sys
is_pypy = '__pypy__' in sys.builtin_module_names

# Needed to make getaddrinfo() work in pymongo on Mac OS X
# Docs mention it's a better choice for Linux as well.
# This must be done asap in the worker
if "GEVENT_RESOLVER" not in os.environ and not is_pypy:
    os.environ["GEVENT_RESOLVER"] = "ares"

from gevent import monkey
monkey.patch_all()

import argparse

sys.path.insert(0, os.getcwd())

from mrq import config
from mrq.agent import Agent
from mrq.context import set_current_config


def main():

    parser = argparse.ArgumentParser(description='Start a MRQ agent')

    cfg = config.get_config(parser=parser, config_type="agent", sources=("file", "env", "args"))

    set_current_config(cfg)

    agent = Agent()

    agent.work()

    sys.exit(agent.exitcode)


if __name__ == "__main__":
    main()



from mrq.task import Task
from mrq.context import connections


class EnsureIndexes(Task):

    def run(self, params):

        if connections.mongodb_logs:
            connections.mongodb_logs.mrq_logs.ensure_index(
                [("job", 1)], background=True)
            connections.mongodb_logs.mrq_logs.ensure_index(
                [("worker", 1)], background=True, sparse=True)

        connections.mongodb_jobs.mrq_workers.ensure_index(
            [("status", 1)], background=True)
        connections.mongodb_jobs.mrq_workers.ensure_index(
            [("datereported", 1)], background=True, expireAfterSeconds=3600)

        connections.mongodb_jobs.mrq_jobs.ensure_index(
            [("status", 1)], background=True)
        connections.mongodb_jobs.mrq_jobs.ensure_index(
            [("path", 1)], background=True)
        connections.mongodb_jobs.mrq_jobs.ensure_index(
            [("worker", 1)], background=True, sparse=True)
        connections.mongodb_jobs.mrq_jobs.ensure_index(
            [("queue", 1)], background=True)
        connections.mongodb_jobs.mrq_jobs.ensure_index(
            [("dateexpires", 1)], sparse=True, background=True, expireAfterSeconds=0)
        connections.mongodb_jobs.mrq_jobs.ensure_index(
            [("dateretry", 1)], sparse=True, background=True)
        connections.mongodb_jobs.mrq_jobs.ensure_index(
            [("datequeued", 1)], background=True)
        connections.mongodb_jobs.mrq_jobs.ensure_index(
            [("queue", 1), ("status", 1), ("datequeued", 1), ("_id", 1)], background=True)
        connections.mongodb_jobs.mrq_jobs.ensure_index(
            [("status", 1), ("queue", 1), ("path", 1)], background=True)

        connections.mongodb_jobs.mrq_scheduled_jobs.ensure_index(
            [("hash", 1)], unique=True, background=False)

        connections.mongodb_jobs.mrq_agents.ensure_index(
            [("datereported", 1)], background=True)
        connections.mongodb_jobs.mrq_agents.ensure_index(
            [("dateexpires", 1)], background=True, expireAfterSeconds=0)
        connections.mongodb_jobs.mrq_agents.ensure_index(
            [("worker_group", 1)], background=True)



from future import standard_library
standard_library.install_aliases()
from future.builtins import str, bytes
from future.utils import iteritems
import gevent
import gevent.pool
import os
import signal
import datetime
import time
import socket
import traceback
import psutil
import sys
import json as json_stdlib
import ujson as json
from bson import ObjectId
from redis.lock import LuaLock
from collections import defaultdict
from mrq.utils import load_class_by_path

from .job import Job
from .exceptions import (TimeoutInterrupt, StopRequested, JobInterrupt, AbortInterrupt,
                         RetryInterrupt, MaxRetriesInterrupt, MaxConcurrencyInterrupt)
from .context import (set_current_worker, set_current_job, get_current_job, get_current_config,
                      connections, enable_greenlet_tracing, run_task, log)
from .queue import Queue
from .utils import MongoJSONEncoder, MovingAverage
from .processes import Process
from .redishelpers import redis_key


class Worker(Process):
    """ Main worker class """

    # Allow easy overloading
    job_class = Job

    # See the doc for valid statuses
    status = "init"

    mongodb_jobs = None
    mongodb_logs = None
    redis = None

    def __init__(self):

        set_current_worker(self)

        if self.config.get("trace_greenlets"):
            enable_greenlet_tracing()

        self.datestarted = datetime.datetime.utcnow()

        self.done_jobs = 0
        self.max_jobs = self.config["max_jobs"]
        self.max_time = datetime.timedelta(seconds=self.config["max_time"]) or None

        self.paused_queues = set()

        self.connected = False  # MongoDB + Redis

        self.process = psutil.Process(os.getpid())
        self.greenlet = gevent.getcurrent()
        self.graceful_stop = None

        self.work_lock = gevent.lock.Semaphore()

        if self.config.get("worker_id"):
            self.id = ObjectId(self.config["worker_id"])
        else:
            self.id = ObjectId()

        if self.config.get("name"):
            self.name = self.config["name"]
        else:
            # Generate a somewhat human-readable name for this worker
            self.name = "%s.%s" % (socket.gethostname().split(".")[0], os.getpid())

        self.pool_size = self.config["greenlets"]
        self.pool_usage_average = MovingAverage((60 / self.config["report_interval"] or 1))

        self.set_logger()

        self.refresh_queues(fatal=True)
        self.queues_with_notify = list({redis_key("notify", q) for q in self.queues if q.use_notify()})
        self.has_subqueues = any([queue.endswith("/") for queue in self.config["queues"]])

        self.log.info(
            "Starting worker on %s queues with %s greenlets" %
            (len(self.queues), self.pool_size)
        )

        self.gevent_pool = gevent.pool.Pool(self.pool_size)

        # Keep references to main greenlets
        self.greenlets = {}

        # TODO by "tag"?
        self._traced_io = {
            "types": defaultdict(float),
            "tasks": defaultdict(float),
            "total": 0
        }

        if self.config["ensure_indexes"]:
            run_task("mrq.basetasks.indexes.EnsureIndexes", {})

    def set_logger(self):
        import logging

        self.log = logging.getLogger(str(self.id))
        logging.basicConfig(format=self.config["log_format"])
        self.log.setLevel(getattr(logging, self.config["log_level"]))
        # No need to send worker logs to mongo?
        # logger_class = load_class_by_path(self.config["logger"])

        # # All mrq handlers must have worker and collection keyword arguments
        # if self.config["logger"].startswith("mrq"):
        #     self.log_handler = logger_class(collection=self.config["mongodb_logs"], worker=str(self.id), **self.config["logger_config"])
        # else:
        #     self.log_handler = logger_class(**self.config["logger_config"])
        # self.log.addHandler(self.log_handler)

    @property
    def config(self):
        return get_current_config()

    def connect(self, force=False):

        if self.connected and not force:
            return

        # Accessing connections attributes will automatically connect
        self.redis = connections.redis
        self.mongodb_jobs = connections.mongodb_jobs
        self.mongodb_logs = connections.mongodb_logs

        self.connected = True

    def greenlet_scheduler(self):

        redis_scheduler_lock_key = "%s:schedulerlock" % get_current_config()["redis_prefix"]
        while True:
            with LuaLock(connections.redis, redis_scheduler_lock_key,
                         timeout=self.config["scheduler_interval"] + 10, blocking=False, thread_local=False):
                self.scheduler.check()

            time.sleep(self.config["scheduler_interval"])

    def greenlet_report(self):
        """ This greenlet always runs in background to update current status
            in MongoDB every N seconds.

            Caution: it might get delayed when doing long blocking operations.
            Should we do this in a thread instead?
         """

        self.report_worker(w=1)
        while True:
            try:
                self.report_worker()
            except Exception as e:  # pylint: disable=broad-except
                self.log.error("When reporting: %s" % e)
            finally:
                time.sleep(self.config["report_interval"])

    def greenlet_logs(self):
        """ This greenlet always runs in background to update current
            logs in MongoDB every 10 seconds.

            Caution: it might get delayed when doing long blocking operations.
            Should we do this in a thread instead?
         """

        while True:
            try:
                self.flush_logs()
            except Exception as e:  # pylint: disable=broad-except
                self.log.error("When flushing logs: %s" % e)
            finally:
                time.sleep(self.config["report_interval"])

    def greenlet_subqueues(self):

        while True:
            self.refresh_queues()
            time.sleep(self.config["subqueues_refresh_interval"])

    def refresh_queues(self, fatal=False):
        """ Updates the list of currently known queues and subqueues """

        try:
            queues = []
            prefixes = [q for q in self.config["queues"] if q.endswith("/")]
            known_subqueues = Queue.all_known(prefixes=prefixes)

            for q in self.config["queues"]:
                queues.append(Queue(q))
                if q.endswith("/"):
                    for subqueue in known_subqueues:
                        if subqueue.startswith(q):
                            queues.append(Queue(subqueue))

            self.queues = queues

        except Exception as e:  # pylint: disable=broad-except
            self.log.error("When refreshing subqueues: %s", e)
            if fatal:
                raise

    def get_paused_queues(self):
        """ Returns the set of currently paused queues """
        return {q.decode("utf-8") for q in self.redis.smembers(redis_key("paused_queues"))}

    def greenlet_paused_queues(self):

      while True:

          # Update the process-local list of paused queues
          self.paused_queues = self.get_paused_queues()
          time.sleep(self.config["paused_queues_refresh_interval"])

    def get_memory(self):
        mmaps = self.process.memory_maps()
        mem = {
            "rss": sum([x.rss for x in mmaps]),
            "swap": sum([getattr(x, 'swap', getattr(x, 'swapped', 0)) for x in mmaps])
        }
        mem["total"] = mem["rss"] + mem["swap"]
        return mem

    def get_worker_report(self, with_memory=False):
        """ Returns a dict containing all the data we can about the current status of the worker and
            its jobs. """

        greenlets = []
        for greenlet in list(self.gevent_pool):
            g = {}
            short_stack = []
            stack = traceback.format_stack(greenlet.gr_frame)
            for s in stack[1:]:
                if "/gevent/hub.py" in s:
                    break
                short_stack.append(s)
            g["stack"] = short_stack

            job = get_current_job(id(greenlet))
            if job:
                job.save()
                if job.data:
                    g["path"] = job.data["path"]
                g["datestarted"] = job.datestarted
                g["id"] = str(job.id)
                g["time"] = getattr(greenlet, "_trace_time", 0)
                g["switches"] = getattr(greenlet, "_trace_switches", None)

                # pylint: disable=protected-access
                if job._current_io is not None:
                    g["io"] = job._current_io

            greenlets.append(g)

        # When faking network latency, all sockets are affected, including OS ones, but
        # we still want reliable reports so this is disabled.
        if (not with_memory) or (self.config["add_network_latency"] != "0" and self.config["add_network_latency"]):
            cpu = {
                "user": 0,
                "system": 0,
                "percent": 0
            }
            mem = {"rss": 0, "swap": 0, "total": 0}
        else:
            cpu_times = self.process.cpu_times()
            cpu = {
                "user": cpu_times.user,
                "system": cpu_times.system,
                "percent": self.process.cpu_percent(0)
            }
            mem = self.get_memory()

        # Avoid sharing passwords or sensitive config!
        whitelisted_config = [
            "max_jobs",
            "max_memory"
            "greenlets",
            "processes",
            "queues",
            "dequeue_strategy",
            "scheduler",
            "name",
            "local_ip",
            "external_ip",
            "agent_id",
            "worker_group"
        ]

        io = None
        if self._traced_io:
            io = {}
            for k, v in iteritems(self._traced_io):
                if k == "total":
                    io[k] = v
                else:
                    io[k] = sorted(list(v.items()), reverse=True, key=lambda x: x[1])

        used_pool_slots = len(self.gevent_pool)
        used_avg = self.pool_usage_average.next(used_pool_slots)

        return {
            "status": self.status,
            "config": {k: v for k, v in iteritems(self.config) if k in whitelisted_config},
            "done_jobs": self.done_jobs,
            "usage_avg": used_avg / self.pool_size,
            "datestarted": self.datestarted,
            "datereported": datetime.datetime.utcnow(),
            "name": self.name,
            "io": io,
            "_id": str(self.id),
            "process": {
                "pid": self.process.pid,
                "cpu": cpu,
                "mem": mem
                # https://code.google.com/p/psutil/wiki/Documentation
                # get_open_files
                # get_connections
                # get_num_ctx_switches
                # get_num_fds
                # get_io_counters
                # get_nice
            },
            "jobs": greenlets
        }

    def report_worker(self, w=0):

        report = self.get_worker_report(with_memory=True)

        if self.config["max_memory"] > 0:
            if report["process"]["mem"]["total"] > (self.config["max_memory"] * 1024 * 1024):
                self.shutdown_max_memory()

        if self.config["report_file"]:
            with open(self.config["report_file"], "wb") as f:
                f.write(bytes(json.dumps(report, ensure_ascii=False), 'utf-8'))  # pylint: disable=no-member

        if "_id" in report:
            del report["_id"]

        try:

            self.mongodb_jobs.mrq_workers.update({
                "_id": ObjectId(self.id)
            }, {"$set": report}, upsert=True, w=w)
        except Exception as e:  # pylint: disable=broad-except
            self.log.debug("Worker report failed: %s" % e)

    def greenlet_timeouts(self):
        """ This greenlet kills jobs in other greenlets if they timeout.
        """

        while True:
            now = datetime.datetime.utcnow()
            for greenlet in list(self.gevent_pool):
                job = get_current_job(id(greenlet))
                if job and job.timeout and job.datestarted:
                    expires = job.datestarted + datetime.timedelta(seconds=job.timeout)
                    if now > expires:
                        job.kill(block=False, reason="timeout")

            time.sleep(1)


    def greenlet_admin(self):
        """ This greenlet is used to get status information about the worker
            when --admin_port was given
        """

        if self.config["processes"] > 1:
            self.log.debug(
                "Admin server disabled because of multiple processes.")
            return

        class Devnull(object):
            def write(self, *_):
                pass

        from gevent import pywsgi

        def admin_routes(env, start_response):
            path = env["PATH_INFO"]
            status = "200 OK"
            res = ""
            if path in ["/", "/report", "/report_mem"]:
                report = self.get_worker_report(with_memory=(path == "/report_mem"))
                res = bytes(json_stdlib.dumps(report, cls=MongoJSONEncoder), 'utf-8')
            elif path == "/wait_for_idle":
                self.wait_for_idle()
                res = bytes("idle", "utf-8")
            else:
                status = "404 Not Found"
            start_response(status, [('Content-Type', 'application/json')])
            return [res]

        server = pywsgi.WSGIServer((self.config["admin_ip"], self.config["admin_port"]), admin_routes, log=Devnull())

        try:
            self.log.debug("Starting admin server on port %s" % self.config["admin_port"])
            server.serve_forever()
        except Exception as e:  # pylint: disable=broad-except
            self.log.debug("Error in admin server : %s" % e)

    def flush_logs(self):
        for handler in self.log.handlers:
            handler.flush()

    def wait_for_idle(self):
        """ Waits until the worker has nothing more to do. Very useful in tests """

        # Be mindful that this is being executed in a different greenlet than the work_* methods.

        while True:

            time.sleep(0.01)

            with self.work_lock:

                if self.status != "wait":
                    continue

                if len(self.gevent_pool) > 0:
                    continue

                # Force a refresh of the current subqueues, one might just have been created.
                self.refresh_queues()

                # We might be dequeueing a new subqueue. Double check that we don't have anything more to do
                outcome, dequeue_jobs = self.work_once(free_pool_slots=1, max_jobs=None)

                if outcome is "wait" and dequeue_jobs == 0:
                    break

    def work(self):
        """Starts the work loop.

        """
        self.work_init()

        self.work_loop(max_jobs=self.max_jobs, max_time=self.max_time)

        self.work_stop()

    def work_init(self):

        self.connect()

        self.status = "started"

        # An interval of 0 disables the refresh
        if self.has_subqueues and self.config["subqueues_refresh_interval"] > 0:
            self.greenlets["subqueues"] = gevent.spawn(self.greenlet_subqueues)

        # An interval of 0 disables the refresh
        if self.config["paused_queues_refresh_interval"] > 0:
            self.greenlets["paused_queues"] = gevent.spawn(self.greenlet_paused_queues)

        if self.config["report_interval"] > 0:
            self.greenlets["report"] = gevent.spawn(self.greenlet_report)
            self.greenlets["logs"] = gevent.spawn(self.greenlet_logs)

        if self.config["admin_port"]:
            self.greenlets["admin"] = gevent.spawn(self.greenlet_admin)

        self.greenlets["timeouts"] = gevent.spawn(self.greenlet_timeouts)

        if self.config["scheduler"] and self.config["scheduler_interval"] > 0:

            from .scheduler import Scheduler
            self.scheduler = Scheduler(self.mongodb_jobs.mrq_scheduled_jobs, self.config.get("scheduler_tasks") or [])

            self.scheduler.check_config_integrity()  # If this fails, we won't dequeue any jobs

            self.greenlets["scheduler"] = gevent.spawn(self.greenlet_scheduler)

        self.install_signal_handlers()

    def work_loop(self, max_jobs=None, max_time=None):

        self.done_jobs = 0
        self.datestarted_work_loop = datetime.datetime.utcnow()
        self.queue_offset = 0

        try:

            max_time_reached = False

            while True:

                if self.graceful_stop:
                    break

                # If the scheduler greenlet is crashed, fail loudly.
                if self.config["scheduler"] and not self.greenlets["scheduler"]:
                    self.exitcode = 1
                    break

                while True:

                    # we put this here to make sure we have a strict limit on max_time
                    if max_time and datetime.datetime.utcnow() - self.datestarted >= max_time:
                        self.log.info("Reached max_time=%s" % max_time.seconds)
                        max_time_reached = True
                        break

                    free_pool_slots = self.gevent_pool.free_count()

                    if max_jobs:
                        total_started = (self.pool_size - free_pool_slots) + self.done_jobs
                        free_pool_slots = min(free_pool_slots, max_jobs - total_started)
                        if free_pool_slots == 0:
                            break

                    if free_pool_slots > 0:
                        break

                    self.status = "full"
                    self.gevent_pool.wait_available(timeout=60)

                if max_time_reached:
                    break

                self.status = "spawn"
                with self.work_lock:
                    outcome, dequeue_jobs = self.work_once(free_pool_slots=free_pool_slots, max_jobs=max_jobs)
                self.status = "wait"

                if outcome == "break":
                    break

                if outcome == "wait":
                    self.work_wait()

        except StopRequested:
            pass

        finally:

            try:

                self.log.debug("Joining the greenlet pool...")
                self.status = "join"

                self.gevent_pool.join(timeout=None, raise_error=False)
                self.log.debug("Joined.")

            except StopRequested:
                pass

        self.datestopped_work_loop = datetime.datetime.utcnow()
        lifetime = self.datestopped_work_loop - self.datestarted_work_loop
        job_rate = float(self.done_jobs) / lifetime.total_seconds()
        self.log.info("Worker spent %.3f seconds performing %s jobs (%.3f jobs/second)" % (
            lifetime.total_seconds(), self.done_jobs, job_rate
        ))

    def work_once(self, free_pool_slots=1, max_jobs=None):
        """ Does one lookup for new jobs, inside the inner work loop """

        dequeued_jobs = 0

        available_queues = [
            queue for queue in self.queues
            if queue.root_id not in self.paused_queues and
            queue.id not in self.paused_queues
        ]

        for queue_i in range(len(available_queues)):

            queue = available_queues[(queue_i + self.queue_offset) % len(available_queues)]

            max_jobs_per_queue = free_pool_slots - dequeued_jobs

            if max_jobs_per_queue <= 0:
                queue_i -= 1
                break

            if self.config["dequeue_strategy"] == "parallel":
                max_jobs_per_queue = max(1, int(max_jobs_per_queue / (len(available_queues) - queue_i)))

            for job in queue.dequeue_jobs(
                max_jobs=max_jobs_per_queue,
                job_class=self.job_class,
                worker=self
            ):
                dequeued_jobs += 1

                self.gevent_pool.spawn(self.perform_job, job)

        # At the next pass, start at the next queue to avoid always dequeuing the same one
        if self.config["dequeue_strategy"] == "parallel":
            self.queue_offset = (self.queue_offset + queue_i + 1) % len(self.queues)

        # TODO consider this when dequeuing jobs to have strict limits
        if max_jobs and self.done_jobs >= max_jobs:
            self.log.info("Reached max_jobs=%s" % self.done_jobs)
            return "break", dequeued_jobs

        # We seem to have exhausted available jobs, we can sleep for a
        # while.
        if dequeued_jobs == 0:

            if self.config["dequeue_strategy"] == "burst":
                self.log.info("Burst mode: stopping now because queues were empty")
                return "break", dequeued_jobs

            return "wait", dequeued_jobs

        return None, dequeued_jobs

    def work_wait(self):
        """ Wait for new jobs to arrive """

        if len(self.queues_with_notify) > 0:
            # https://github.com/antirez/redis/issues/874
            connections.redis.blpop(*(self.queues_with_notify + [max(1, int(self.config["max_latency"]))]))
        else:
            gevent.sleep(self.config["max_latency"])

    def work_stop(self):

        self.status = "kill"

        self.gevent_pool.kill(exception=JobInterrupt, block=True)

        for g in self.greenlets:
            g_time = getattr(self.greenlets[g], "_trace_time", 0)
            g_switches = getattr(self.greenlets[g], "_trace_switches", None)
            self.greenlets[g].kill(block=True)
            self.log.debug(
                "Greenlet for %s killed (%0.5fs, %s switches)." %
                (g, g_time, g_switches))

        self.status = "stop"

        self.report_worker(w=1)
        self.flush_logs()

        g_time = getattr(self.greenlet, "_trace_time", 0)
        g_switches = getattr(self.greenlet, "_trace_switches", None)
        self.log.debug(
            "Exiting main worker greenlet (%0.5fs, %s switches)." %
            (g_time, g_switches))

    def perform_job(self, job):
        """ Wraps a job.perform() call with timeout logic and exception handlers.

            This is the first call happening inside the greenlet.
        """

        if self.config["trace_memory"]:
            job.trace_memory_start()

        set_current_job(job)

        extra = {"mrq_job_id": str(job.id), "data": job.data}

        try:
            job.perform()

        except MaxConcurrencyInterrupt as e:
            self.log.exception("MRQ Max concurrency reached: %s" % str(job.data['path']), extra = extra)
            job._save_status("maxconcurrency", exception=True)

        except RetryInterrupt as e:
            self.log.exception("MRQ Caught retry: %s" % str(job.data['path']), extra = extra)
            job.save_retry(sys.exc_info()[1])

        except MaxRetriesInterrupt as e:
            self.log.exception("MRQ Max retries reached: %s" % str(job.data['path']), extra = extra)
            job._save_status("maxretries", exception=True)

        except AbortInterrupt as e:
            self.log.exception("MRQ Caught abort: %s" % str(job.data['path']), extra = extra)
            job.save_abort()

        except TimeoutInterrupt:
            self.log.exception("MRQ %s timeouted after %s seconds" % (str(job.data['path']), job.timeout), extra = extra)
            job._save_status("timeout", exception=True)

        except JobInterrupt as e:
            self.log.exception("MRQ Job interrupted: %s" % str(job.data['path']), extra = extra)
            job._save_status("interrupt", exception=True)

        except Exception as e:
            self.log.exception("MRQ Job failed: %s" % str(job.data['path']), extra = extra)
            job._save_status("failed", exception=True)

        finally:

            set_current_job(None)

            self.done_jobs += 1

            if self.config["trace_memory"]:
                job.trace_memory_stop()

    def shutdown_graceful(self):
        """ Graceful shutdown: waits for all the jobs to finish. """

        self.log.info("Graceful shutdown...")
        raise StopRequested()  # pylint: disable=nonstandard-exception

    def shutdown_max_memory(self):

        self.log.info("Max memory reached, shutdown...")
        self.graceful_stop = True

    def shutdown_now(self):
        """ Forced shutdown: interrupts all the jobs. """

        self.log.info("Forced shutdown...")
        self.status = "killing"

        self.gevent_pool.kill(exception=JobInterrupt, block=False)

        raise StopRequested()  # pylint: disable=nonstandard-exception



from future import standard_library
standard_library.install_aliases()
from past.builtins import basestring
from .context import get_current_job, get_current_worker
import time
import random
import re
import copy


def patch_method(base_class, method_name, method):

    # Create a closure to store the old method
    def _patched_factory(old_method):
        def _mrq_patched_method(*args, **kwargs):
            return method(old_method, *args, **kwargs)
        return _mrq_patched_method

    old_method = getattr(base_class, method_name)
    setattr(base_class, method_name, _patched_factory(old_method))


def patch_io_all(config):
    """ Patch higher-level modules to provide insights on what is performing IO for each job. """

    patch_io_httplib(config)
    patch_io_redis(config)
    patch_io_pymongo_cursor(config)

    #patch_io_dns()
    #patch_io_subprocess()


def patch_pymongo(config):
    """ Monkey-patch pymongo's collections to add some logging """

    # Nothing to change!
    if not config["print_mongodb"] and not config["trace_io"]:
        return

    from termcolor import cprint

    # Print because we are very early and log() may not be ready yet.
    cprint("Monkey-patching MongoDB methods...", "white")

    def gen_monkey_patch(base_object, method):
        base_method = getattr(base_object, method)

        def mrq_monkey_patched(self, *args, **kwargs):

            if config["trace_io"]:
                comment = "mrq"

                worker = get_current_worker()
                job = get_current_job()
                if job:
                    job.set_current_io({
                        "type": "mongodb.%s" % method,
                        "data": {
                            "collection": self.full_name
                        }
                        # Perf issue? All MongoDB data will get jsonified!
                        # "data": json.dumps(args)[0:300]
                    })
                    comment = {"job": job.id}
                elif worker:
                    comment = {"worker": worker.id}

                # Tag potentially expensive queries with their job id for easier debugging
                if method in ["find", "find_and_modify", "count", "update_many", "update", "delete_many"]:
                    if len(args) > 0 and isinstance(args[0], dict) and "$comment" not in args[0]:
                        query = copy.copy(args[0])
                        query["$comment"] = comment
                        args = (query, ) + args[1:]

            if config["print_mongodb"]:
                if self.full_name in config.get("print_mongodb_hidden_collections", []):
                    cprint("[MONGO] %s.%s%s %s" % (
                        self.full_name, method, "-hidden-", kwargs
                    ), "magenta")
                else:
                    cprint("[MONGO] %s.%s%s %s" % (
                        self.full_name, method, args, kwargs
                    ), "magenta")

            if config.get("mongodb_pre_hook"):
                config.get("mongodb_pre_hook")({
                    "collection": self.full_name,
                    "method": method,
                    "args": args,
                    "kwargs": kwargs,
                    "client": self.database.client,
                    "job": job
                })

            start_time = time.time()
            ret = False
            try:
                ret = base_method(self, *args, **kwargs)
            finally:
                stop_time = time.time()

                job = None

                if config["trace_io"]:
                    job = get_current_job()
                    if job:
                        job.set_current_io(None)

                if config.get("mongodb_post_hook"):
                    config.get("mongodb_post_hook")({
                        "collection": self.full_name,
                        "method": method,
                        "args": args,
                        "kwargs": kwargs,
                        "client": self.database.client,
                        "job": job,
                        "result": ret,
                        "time": stop_time - start_time
                    })

            return ret

        # Needed to avoid breaking mongokit
        mrq_monkey_patched.__doc__ = method.__doc__

        return mrq_monkey_patched

    from pymongo.collection import Collection
    pymongo_method_whitelist = (
        "bulk_write",
        "find", "find_one_and_delete", "find_one_and_replace", "find_one_and_update",
        "update", "update_one", "update_many",
        "drop",
        "count",
        "save",
        "insert", "insert_one", "insert_many",
        "replace_one",
        "remove", "delete_one", "delete_many",
        "find_and_modify",
        "parallel_scan",
        "options",
        "aggregate",
        "group", "distinct",
        "rename",
        "map_reduce", "inline_map_reduce",
        "create_indexes", "create_index", "ensure_index", "drop_index", "reindex", "list_indexes"
    )
    for method in pymongo_method_whitelist:
        if hasattr(Collection, method) and getattr(Collection, method).__name__ != "mrq_monkey_patched":
            setattr(Collection, method, gen_monkey_patch(Collection, method))


def patch_network_latency(seconds=0.01):
    """ Add random latency to all I/O operations """

    # Accept float(0.1), "0.1", "0.1-0.2"
    def sleep():
        if isinstance(seconds, float):
            time.sleep(seconds)
        elif isinstance(seconds, basestring):
            # pylint: disable=maybe-no-member
            if "-" in seconds:
                time.sleep(random.uniform(
                    float(seconds.split("-")[0]),
                    float(seconds.split("-")[1])
                ))
            else:
                time.sleep(float(seconds))

    def _patched_method(old_method, *args, **kwargs):
        sleep()
        return old_method(*args, **kwargs)

    socket_methods = [
        "send", "sendall", "sendto", "recv", "recvfrom", "recvfrom_into", "recv_into",
        "connect", "connect_ex", "close"
    ]

    from socket import socket as _socketmodule
    from gevent.socket import socket as _geventmodule
    from gevent.ssl import SSLSocket as _sslmodule   # pylint: disable=no-name-in-module

    for method in socket_methods:
        patch_method(_socketmodule, method, _patched_method)
        patch_method(_geventmodule, method, _patched_method)
        patch_method(_sslmodule, method, _patched_method)


def patch_io_redis(config):

    def execute_command(old_method, self, *args, **options):

        job = get_current_job()
        if job:
            job.set_current_io({
                "type": "redis.%s" % args[0].lower(),
                "data": {
                    "key": args[1] if len(args) > 1 else None
                }
            })

        if config.get("redis_pre_hook"):
            config.get("redis_pre_hook")({
                "command": args[0],
                "args": args[1:],
                "options": options,
                "client": self,
                "job": job
            })

        start_time = time.time()
        ret = False
        try:
            ret = old_method(self, *args, **options)
        finally:
            stop_time = time.time()

            if job:
                job.set_current_io(None)

            if config.get("redis_post_hook"):
                config.get("redis_post_hook")({
                    "command": args[0],
                    "args": args[1:],
                    "options": options,
                    "time": stop_time - start_time,
                    "result": ret,
                    "job": job,
                    "client": self
                })

        return ret

    from redis import StrictRedis

    patch_method(StrictRedis, "execute_command", execute_command)


def patch_io_httplib(config):
    """ Patch the base httplib.HTTPConnection class, which is used in most HTTP libraries
        like urllib2 or urllib3/requests. """

    # pylint: disable=import-error

    def start(method, url):
        job = get_current_job()
        if job:
            job.set_current_io({
                "type": "http.%s" % method.lower(),
                "data": {
                    "url": url
                }
            })

    def stop():
        job = get_current_job()
        if job:
            job.set_current_io(None)

    class mrq_wrapped_socket(object):
        """ Socket-like object that keeps track of 'trace_args' and wraps our monitoring code
            around blocking I/O calls. """

        def __init__(self, obj, parent_connection):
            self._obj = obj
            self._parent_connection = parent_connection

            def _make_patched_method(method):
                def _patched_method(*args, **kwargs):

                    # In the case of HTTPS, we may connect() before having called conn.request()
                    # For requests/urllib3, we may need to plug ourselves at the
                    # connectionpool.urlopen level
                    if not hasattr(self._parent_connection, "_traced_args"):
                        return getattr(self._obj, method)(*args, **kwargs)

                    start(*self._parent_connection._traced_args)  # pylint: disable=protected-access
                    try:
                        data = getattr(self._obj, method)(*args, **kwargs)
                    finally:
                        stop()
                    return data
                return _patched_method

            # Replace socket methods with instrumented ones
            for method in [

              # socket
              "send", "sendall", "sendto", "recv", "recvfrom", "recvfrom_into", "recv_into",
              "connect", "connect_ex", "close",

              # fileobject
              "read", "readline", "write", "writelines", "seek"
            ]:
                setattr(self, method, _make_patched_method(method))

        # Forward all other calls/attributes to the base socket
        def __getattr__(self, attr):
            # cprint(attr, "green")
            return getattr(self._obj, attr)

        def makefile(self, *args, **kwargs):
            newsock = self._obj.makefile(*args, **kwargs)
            return mrq_wrapped_socket(newsock, self._parent_connection)

    def request(old_method, self, method, url, body=None, headers=None, *args, **kwargs):

        if headers is None:
            headers = {}

        # This is for proxy support - TODO show that in dashboard?
        if re.search(r"^http(s?)\:\/\/", url):
            report_url = url
        else:
            protocol = "http"
            if hasattr(self, "key_file"):
                protocol = "https"

            report_url = "%s://%s%s%s" % (
                protocol,
                self.host,
                (":%s" % self.port) if self.port != 80 else "",
                url
            )

        self._traced_args = (method, report_url)  # pylint: disable=protected-access
        res = old_method(self, method, url, body=body, headers=headers)
        return res

    def connect(old_method, self, *args, **kwargs):

        # In the case of HTTPS, we may connect() before having called conn.request()
        # For requests/urllib3, we may need to plug ourselves at the connectionpool.urlopen level
        if not hasattr(self, "_traced_args"):
            ret = old_method(self, *args, **kwargs)
        else:
            start(*self._traced_args)  # pylint: disable=protected-access
            try:
                ret = old_method(self, *args, **kwargs)
            finally:
                stop()
        self.sock = mrq_wrapped_socket(self.sock, self)

        return ret

    from http.client import HTTPConnection, HTTPSConnection

    patch_method(HTTPConnection, "request", request)
    patch_method(HTTPConnection, "connect", connect)
    patch_method(HTTPSConnection, "connect", connect)

    # Try to patch requests & urllib3 as they are very popular python modules.
    try:
        from requests.packages.urllib3.connection import (
            HTTPConnection,
            UnverifiedHTTPSConnection,
            VerifiedHTTPSConnection
        )

        patch_method(HTTPConnection, "connect", connect)
        patch_method(UnverifiedHTTPSConnection, "connect", connect)
        patch_method(VerifiedHTTPSConnection, "connect", connect)

    except ImportError:
        pass

    try:
        from urllib3.connection import (
            HTTPConnection,
            UnverifiedHTTPSConnection,
            VerifiedHTTPSConnection
        )

        patch_method(HTTPConnection, "connect", connect)
        patch_method(UnverifiedHTTPSConnection, "connect", connect)
        patch_method(VerifiedHTTPSConnection, "connect", connect)

    except ImportError:
        pass


def patch_io_pymongo_cursor(config):

    from pymongo.cursor import Cursor

    class mrq_patched_pymongo_cursor(Cursor):

        # Some dark magic is needed here to cope with python's name mangling for private variables.
        def _Cursor__send_message(self, *args, **kwargs):

            subtype = "cursor"
            collection = self._Cursor__collection.name  # pylint: disable=no-member

            if collection == "$cmd":
                items = list(self._Cursor__spec.items())  # pylint: disable=no-member
                if len(items) > 0:
                    subtype, collection = items[0]

            full_name = "%s.%s" % (self._Cursor__collection.database.name, collection)  # pylint: disable=no-member

            job = get_current_job()

            if job:

                job.set_current_io({
                    "type": "mongodb.%s" % subtype,
                    "data": {
                        "collection": full_name
                    }
                })

            if config.get("mongodb_pre_hook"):

                config.get("mongodb_pre_hook")({
                    "collection": full_name,
                    "method": subtype,
                    "args": (getattr(args[0], "spec", None), ),
                    "kwargs": kwargs,
                    "client": self._Cursor__collection.database.client,
                    "job": job
                })

            start_time = time.time()
            ret = False
            try:
                ret = Cursor._Cursor__send_message(self, *args, **kwargs)  # pylint: disable=no-member
            finally:
                stop_time = time.time()
                if job:
                    job.set_current_io(None)

                if config.get("mongodb_post_hook"):
                    config.get("mongodb_post_hook")({
                        "collection": full_name,
                        "method": subtype,
                        "args": (getattr(args[0], "spec", None), ),
                        "kwargs": kwargs,
                        "client": self._Cursor__collection.database.client,
                        "job": job,
                        "result": ret,
                        "time": stop_time - start_time
                    })

            return ret

    import pymongo as pymongomodule
    if pymongomodule.cursor.Cursor.__name__ != "mrq_patched_pymongo_cursor":
        pymongomodule.cursor.Cursor = mrq_patched_pymongo_cursor

    if pymongomodule.collection.Cursor.__name__ != "mrq_patched_pymongo_cursor":
        pymongomodule.collection.Cursor = mrq_patched_pymongo_cursor
